package org.processmining.dataawarereplayer.precision;

import java.text.NumberFormat;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.TimeUnit;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.dataawarereplayer.precision.measure.PrecisionMeasure.ResultMode;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjectionException;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjector;
import org.processmining.dataawarereplayer.precision.projection.ProcessProjectorAlignmentImpl;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.ui.ConfigurationUIHelper;
import org.processmining.framework.connections.Connection;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.annotations.ConnectionObjectFactory;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.PluginExecutionResult;
import org.processmining.framework.plugin.PluginParameterBinding;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.Pair;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.GUI.MappingPanel;
import org.processmining.plugins.balancedconformance.AbstractBalancedDataConformancePlugin;
import org.processmining.plugins.balancedconformance.BalancedDataXAlignmentPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.result.AlignmentCollection;
import org.processmining.plugins.balancedconformance.ui.DataConformanceConfigUIUtils;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignedLog;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;

import com.google.common.base.Splitter;
import com.google.common.base.Stopwatch;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.SetMultimap;

/**
 * Plugin implementing a method to measure the precision of multi-perspective
 * process models with regards to an event log. The technique builds on
 * {@link DataPetriNet} and is described in the BPI'15/LNBIP paper 'Measuring
 * the Precision of Multi-perspective Process Models'.
 * 
 * @author F. Mannhardt
 *
 */
@Plugin(name = "Precision of DPN", level = PluginLevel.PeerReviewed, parameterLabels = { "Data Petri Net",
		"Log" }, returnLabels = { "Multi-perspective Activity Precision Measurement" }, //
		returnTypes = { PrecisionResult.class }, userAccessible = true, categories = { PluginCategory.Analytics,
				PluginCategory.ConformanceChecking }, keywords = "precision,dpn", help = "Calculate the data-aware precision of a Data Petri Net (DPN) given an event log."
						+ "The technique is described in the BPI'15 paper 'Measuring the Precision of Multi-perspective Process Models' (DOI: 10.1007/978-3-319-42887-1_10)")
public final class DataAwarePrecisionPlugin {

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	@PluginVariant(variantLabel = "Precision of DPN", requiredParameterLabels = { 0, 1 })
	public PrecisionResult measurePrecision(UIPluginContext context, Petrinet graph, XLog log)
			throws UserCancelledException {
		return measurePrecision(context, DataPetriNet.Factory.viewAsDataPetriNet(graph), log);
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	@PluginVariant(variantLabel = "Precision of DPN", requiredParameterLabels = { 0, 1 })
	public PrecisionResult measurePrecision(UIPluginContext context, DataPetriNet net, XLog log)
			throws UserCancelledException {

		BalancedDataXAlignmentPlugin dataConformance = new BalancedDataXAlignmentPlugin();
		BalancedProcessorConfiguration config = AbstractBalancedDataConformancePlugin.queryConfiguration(context, net,
				log);

		config.setVariablesUnassignedMode(UnassignedMode.NULL);

		try {
			XLog alignedLog = dataConformance.alignLog(context.getProgress(), net, log, config);
			ProvidedObjectHelper.publish(context, "Data Alignment Result " + net.getLabel(), alignedLog, XLog.class,
					false);

			XAlignedLog extendLog = XAlignmentExtension.instance().extendLog(alignedLog);

			PrecisionConfig precisionConfig = new PrecisionConfig(getInitialMarking(context, net),
					convertMapping(config.getActivityMapping()), config.getActivityMapping().getEventClassifier(),
					config.getVariableMapping());

			queryPrecisionConfig(context, net, log, precisionConfig);

			Stopwatch stopWatch = Stopwatch.createStarted();
			PrecisionResult result = doMeasurePrecisionWithAlignment(net, log, extendLog, precisionConfig);
			context.log("Computed precision in: "
					+ NumberFormat.getIntegerInstance().format(stopWatch.elapsed(TimeUnit.MILLISECONDS)));

			return result;

		} catch (ControlFlowAlignmentException | DataAlignmentException e) {
			context.log(e);
			ProMUIHelper.showErrorMessage(context.getGlobalContext().getUI(), e.getMessage(), "Alignment failed",
					e);
			context.getFutureResult(0).cancel(true);
			return null;
		} catch (PrecisionMeasureException | ProcessProjectionException e) {
			context.log(e);
			ProMUIHelper.showErrorMessage(context.getGlobalContext().getUI(), e.getMessage(), "Precision Measurement Failed",
					e);
			context.getFutureResult(0).cancel(true);
			return null;
		}
	}

	@Plugin(name = "Precision of DPN (Use Aligned Log)", level = PluginLevel.Regular, parameterLabels = {
			"Data Petri Net", "Log", "Aligned Log (XAlignmentExtension)" }, returnLabels = {
					"Multi-perspective Activity Precision Measurement" }, //
			returnTypes = { PrecisionResult.class }, userAccessible = true, categories = { PluginCategory.Analytics,
					PluginCategory.ConformanceChecking }, keywords = "precision,dpn", help = "Calculate the data-aware precision of a Data Petri Net (DPN) given an event log and an aligned log using the XAlignmentExtension."
							+ "The technique is described in the BPI'15 paper 'Measuring the Precision of Multi-perspective Process Models' (DOI: 10.1007/978-3-319-42887-1_10)")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public PrecisionResult measurePrecisionOfPetrinetWithAlignment(UIPluginContext context, Petrinet graph, XLog log,
			XLog alignedLog) throws PrecisionMeasureException, ProcessProjectionException, UserCancelledException {
		return measurePrecisionWithAlignment(context, DataPetriNet.Factory.viewAsDataPetriNet(graph), log, alignedLog);
	}

	@Plugin(name = "Precision of DPN (Use Aligned Log)", level = PluginLevel.Regular, parameterLabels = {
			"Data Petri Net", "Log", "Aligned Log (XAlignmentExtension)" }, returnLabels = {
					"Multi-perspective Activity Precision Measurement" }, //
			returnTypes = { PrecisionResult.class }, userAccessible = true, categories = { PluginCategory.Analytics,
					PluginCategory.ConformanceChecking }, keywords = "precision,dpn", help = "Calculate the data-aware precision of a Data Petri Net (DPN) given an event log and an aligned log using the XAlignmentExtension."
							+ "The technique is described in the BPI'15 paper 'Measuring the Precision of Multi-perspective Process Models' (DOI: 10.1007/978-3-319-42887-1_10)")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public PrecisionResult measurePrecisionWithAlignment(UIPluginContext context, DataPetriNet net, XLog log,
			XLog alignedLog) throws PrecisionMeasureException, ProcessProjectionException, UserCancelledException {

		TransEvClassMapping eventMapping = ConfigurationUIHelper.queryActivityEventClassMapping(context, net, log);
		MappingPanel<String, String> mappingPanel = DataConformanceConfigUIUtils.queryVariableMapping(context, net,
				XUtils.getEventAttributeKeys(log));

		PrecisionConfig config = new PrecisionConfig(getInitialMarking(context, net), convertMapping(eventMapping),
				eventMapping.getEventClassifier(), mappingPanel.getMapping(false));

		queryPrecisionConfig(context, net, log, config);

		XAlignedLog xAlignedLog = XAlignmentExtension.instance().extendLog(alignedLog);

		Stopwatch stopWatch = Stopwatch.createStarted();
		PrecisionResult result = doMeasurePrecisionWithAlignment(net, log, xAlignedLog, config);
		context.log("Computation Time Precision: "
				+ NumberFormat.getIntegerInstance().format(stopWatch.elapsed(TimeUnit.MILLISECONDS)));

		return result;
	}

	/**
	 * Headless method using an already calculated {@link AlignmentCollection}.
	 * 
	 * @param net
	 * @param log
	 * @param alignedLog
	 * @param precisionConfig
	 * @return
	 * @throws PrecisionMeasureException
	 * @throws ProcessProjectionException
	 */
	public PrecisionResult doMeasurePrecisionWithAlignment(DataPetriNet net, XLog log, Iterable<XAlignment> alignedLog,
			PrecisionConfig precisionConfig) throws PrecisionMeasureException, ProcessProjectionException {
		return doMeasurePrecisionWithAlignment(net, log, alignedLog, precisionConfig,
				XUtils.createEventClasses(precisionConfig.getClassifier(), log));
	}

	/**
	 * 
	 * Headless method using an already calculated {@link AlignmentCollection}.
	 * 
	 * @param net
	 * @param log
	 * @param alignedLog
	 * @param precisionConfig
	 * @param eventClasses
	 *            matching the classifier in precisionConfig
	 * @return
	 * @throws PrecisionMeasureException
	 * @throws ProcessProjectionException
	 */
	public PrecisionResult doMeasurePrecisionWithAlignment(DataPetriNet net, XLog log, Iterable<XAlignment> alignedLog,
			PrecisionConfig precisionConfig, XEventClasses eventClasses)
			throws PrecisionMeasureException, ProcessProjectionException {
		return doMeasurePrecision(net, log,
				new ProcessProjectorAlignmentImpl(net, log, alignedLog, eventClasses,
						precisionConfig.getActivityMapping(), precisionConfig.getVariableMapping(),
						precisionConfig.getExtraAttributes(), precisionConfig.getExecutors()),
				precisionConfig);
	}

	/**
	 * Low level method using the specified {@link ProcessProjector} to get the
	 * aligned traces. This method does not create connections.
	 * 
	 * @param net
	 * @param log
	 * @param processProjector
	 * @param config
	 * @return
	 * @throws PrecisionMeasureException
	 */
	public PrecisionResult doMeasurePrecision(DataPetriNet net, XLog log, ProcessProjector processProjector,
			PrecisionConfig config) throws PrecisionMeasureException {
		DataAwarePrecision dataAwarePrecision = new DataAwarePrecision(config, net, log, processProjector);
		PrecisionResult precisionResult = dataAwarePrecision.computeDataAwarePrecision(ResultMode.PLACE_LEVEL);
		return precisionResult;
	}

	public static void queryPrecisionConfig(UIPluginContext context, DataPetriNet net, XLog log, PrecisionConfig config)
			throws UserCancelledException {

		SetMultimap<String, Transition> eventIdTransitionMap = config.getActivityMapping();
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, config.getClassifier());

		@SuppressWarnings("rawtypes")
		Map<String, Class> attributeType = new HashMap<>();
		Set<String> attributeInLog = new TreeSet<>();
		SetMultimap<Transition, String> activityExecutorMap = HashMultimap.create();

		for (XTrace trace : log) {
			for (XEvent e : trace) {

				Set<Transition> mappedTransitions = eventIdTransitionMap
						.get(logInfo.getEventClasses().getClassOf(e).getId());

				for (XAttribute a : e.getAttributes().values()) {
					if (!isSpecialAttribute(a)) {
						attributeInLog.add(a.getKey());
						//TODO detect inconsistent classes
						attributeType.put(a.getKey(), XUtils.getAttributeValue(a).getClass());
					}
					String resource = XOrganizationalExtension.instance().extractResource(e);
					if (resource != null) {
						for (Transition transition : mappedTransitions) {
							activityExecutorMap.put(transition, resource);
						}
					}
				}

			}
		}

		Set<String> attributesInNet = new HashSet<>();
		for (DataElement element : net.getVariables()) {
			String attributeName = config.getVariableMapping().get(element.getVarName());
			if (attributeName != null) {
				attributesInNet.add(attributeName);
				attributeInLog.remove(attributeName); // already considered through variable mapping
			} else {
				context.log("Unmapped variable " + element.toString());
			}
		}

		PrecisionConfigPanel configPanel = new PrecisionConfigPanel(net.getTransitions(), activityExecutorMap,
				attributeInLog);

		InteractionResult precisionConfigResult = context.showConfiguration("Parameters for DPN-Precision",
				configPanel);
		if (precisionConfigResult == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			config.setDebug(configPanel.getIsDebugMode());
			config.setPenalizeImpreciseEnding(configPanel.getIsPenalizeImpreciseEndings());
			config.setAssumeUnassignedVariablesFree(configPanel.getIsAssumeUnassignedVariablesFree());
			config.setConcurrentThreads(configPanel.getConcurrentThreads());
			Set<String> relevantAttributes = configPanel.getRelevantAttributes();
			// Minus the already mapped DPN-net variables
			relevantAttributes.removeAll(config.getVariableMapping().values());
			@SuppressWarnings("rawtypes")
			Map<String, Class> extraAttributes = new HashMap<>();
			for (String attribute : relevantAttributes) {
				extraAttributes.put(attribute, attributeType.get(attribute));
			}
			config.setExtraAttributes(extraAttributes);
			config.setExecutors(convertExecutors(configPanel.getExecutors()));
			config.setPrecisionMeasure(configPanel.getPrecisionMeasure());
		}

	}

	private static SetMultimap<Transition, String> convertExecutors(Map<Transition, String> map) {
		SetMultimap<Transition, String> executorMaps = HashMultimap.create();
		for (Entry<Transition, String> entry : map.entrySet()) {
			List<String> executors = Splitter.on(',').trimResults().omitEmptyStrings().splitToList(entry.getValue());
			executorMaps.putAll(entry.getKey(), executors);
		}
		return executorMaps;
	}

	private static boolean isSpecialAttribute(XAttribute a) {
		return XUtils.isStandardExtensionAttribute(a);
	}

	public static SetMultimap<String, Transition> convertMapping(TransEvClassMapping mapping) {
		SetMultimap<String, Transition> newEventMapping = HashMultimap.create();
		for (Entry<Transition, XEventClass> entry : mapping.entrySet()) {
			newEventMapping.put(entry.getValue().getId(), entry.getKey());
		}
		return newEventMapping;
	}

	public static Marking getInitialMarking(UIPluginContext context, DataPetriNet net) {
		// check existence of initial marking		
		try {
			InitialMarkingConnection initCon = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, net);

			if (((Marking) initCon.getObjectWithRole(InitialMarkingConnection.MARKING)).isEmpty()) {
				JOptionPane.showMessageDialog(new JPanel(),
						"The initial marking is an empty marking. If this is not intended, remove the currently existing InitialMarkingConnection object and then use \"Create Initial Marking\" plugin to create a non-empty initial marking.",
						"Empty Initial Marking", JOptionPane.INFORMATION_MESSAGE);
			}
		} catch (ConnectionCannotBeObtained exc) {
			if (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(new JPanel(),
					"No initial marking is found for this model. Do you want to create one?", "No Initial Marking",
					JOptionPane.YES_NO_OPTION, JOptionPane.INFORMATION_MESSAGE)) {
				createMarking(context, net, InitialMarkingConnection.class);
			}
		}

		// check connection between petri net and marking
		Marking initMarking = null;
		try {
			initMarking = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, net)
					.getObjectWithRole(InitialMarkingConnection.MARKING);
		} catch (ConnectionCannotBeObtained exc) {
			initMarking = new Marking();
		}
		return initMarking;
	}

	private static boolean createMarking(UIPluginContext context, PetrinetGraph net,
			Class<? extends Connection> classType) {
		boolean result = false;
		Collection<Pair<Integer, PluginParameterBinding>> plugins = context.getPluginManager()
				.find(ConnectionObjectFactory.class, classType, context.getClass(), true, false, false, net.getClass());
		PluginContext c2 = context.createChildContext("Creating connection of Type " + classType);
		Pair<Integer, PluginParameterBinding> pair = plugins.iterator().next();
		PluginParameterBinding binding = pair.getSecond();
		try {
			PluginExecutionResult pluginResult = binding.invoke(c2, net);
			pluginResult.synchronize();
			context.getProvidedObjectManager().createProvidedObjects(c2); // push the objects to main context
			result = true;
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			c2.getParentContext().deleteChild(c2);
		}
		return result;
	}

}
