package org.processmining.plugins.DataConformance;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.plugins.DataConformance.framework.ActivityMatchCosts;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.framework.ReplayState;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.balancedconformance.result.AlignmentCollection;

import com.google.common.collect.Maps;

public class ResultReplay implements AlignmentCollection {

	public interface StatisticsCallback {

		void updateStatistics(ExecutionStep logStep, ExecutionStep processStep);

	}

	/**
	 * The collection of all existing alignments
	 */
	public NavigableSet<Alignment> labelStepArray;

	/**
	 * A map for fast access of alignments by the trace name
	 */
	protected Map<String, Alignment> nameToAlignmentMap;

	/**
	 * It associates an attribute to a 4-element tuple: - the first element is
	 * the number of times the attribute has correctly been written - the second
	 * element is the number of times the attribute hasn't been written - the
	 * third is the average of the misconformance on the data attribute values.
	 * - the fourth is the standard deviation of the misconformance on the data
	 * attribute values. - the fifth is the number of times the attribute has
	 * been written a wrong value
	 */
	public Map<String, float[]> attrArray = new HashMap<>();

	/**
	 * It associates an activity to a 2-element tuple: - the first element is
	 * the number of times an activity is involved in a move a both with the
	 * same attribute assignments; - the second element is the number of times
	 * an activity is involved in a move a both with the different attribute
	 * assignments. - the third element is the number of times an activity is
	 * involved in a move in log only. - the fourth element is the number of
	 * times an activity is involved in a move in model only.
	 */
	public Map<String, float[]> actArray = new HashMap<>();

	public float meanFitness;

	protected VariableMatchCosts variableCost;
	protected float minCost;
	protected ActivityMatchCosts<XEventClass> activityCost;
	
	protected float meanVisitedNode;
	protected float meanBuiltTreeSize;
	private final XLog alignedLog;

	/**
	 * Map from variable name in DPN-net to variable name in event log
	 */
	private final Map<String, String> variableMapping;

	private final XEventClassifier classifier;

	public ResultReplay(Collection<? extends ReplayState> list, ActivityMatchCosts<XEventClass> activityCost,
			VariableMatchCosts variableCost, Map<String, String> variableMapping, XLog alignedLog, XEventClassifier classifier) {
		this.variableMapping = variableMapping;
		this.alignedLog = alignedLog;
		this.variableCost = variableCost;
		this.activityCost = activityCost;
		this.classifier = classifier;
		this.labelStepArray = new TreeSet<>();
		this.nameToAlignmentMap = Maps.newHashMapWithExpectedSize(list.size());
		if (activityCost != null && variableCost != null)
			minCost = Math.min(activityCost.lowestCost(), variableCost.lowestCost());
		calcReplayResult(list);
	}

	protected void calcReplayResult(Collection<? extends ReplayState> list) {
		int traceIndex = -1;

		float sumFitness = 0;
		long visitedNodes = 0;
		long treeSizes = 0;

		for (ReplayState state : list) {
			traceIndex++;
			visitedNodes += state.visitedNodes;
			treeSizes += state.treeSize;
			float fitness = 0;
			fitness = computeFitness(state);
			sumFitness += fitness;
			Alignment align = createAlignment(state, fitness, new StatisticsCallback() {

				public void updateStatistics(ExecutionStep logStep, ExecutionStep processStep) {
					updateStatisticsMap(logStep, processStep);

				}
			});
			labelStepArray.add(align);
			nameToAlignmentMap.put(state.getTraceName(), align);
		}

		meanVisitedNode = visitedNodes / ((float) list.size());
		meanBuiltTreeSize = treeSizes / ((float) list.size());
		meanFitness = sumFitness / (traceIndex + 1);

		for (float[] elem : attrArray.values()) {
			elem[3] = (float) Math.sqrt(elem[3] / (elem[0] - elem[1]));
		}
	}

	/**
	 * Creates an {@link Alignment} object from a {@link ReplayState}, executing
	 * the {@link StatisticsCallback} for each {@link ExecutionStep} of the
	 * alignment.
	 * 
	 * @param state the alignment result
	 * @param fitness the fitness
	 * @param callback may be NULL
	 * @return
	 */
	public static Alignment createAlignment(ReplayState state, float fitness, StatisticsCallback callback) {
		Iterator<ExecutionStep> logTraceIter = state.getLogTracePrefix().iterator();
		Iterator<ExecutionStep> processTraceIter = state.getProcessTracePrefix().iterator();

		List<String> labelArray = new ArrayList<>(state.getLogTracePrefix().size());
		List<DataAwareStepTypes> stepArray = new ArrayList<>();
		Alignment align = new Alignment(state, fitness, labelArray, stepArray);

		while (logTraceIter.hasNext() && processTraceIter.hasNext()) {
			ExecutionStep processStep = processTraceIter.next();
			ExecutionStep logStep = logTraceIter.next();
			
			if (fitness > 0 && callback != null) {
				callback.updateStatistics(logStep, processStep);
			}

			if (logStep.equals(ExecutionStep.bottomStep)) {
				labelArray.add("P: " + processStep.toString() + "  L: -");
				if (processStep.isInvisible()) {
					stepArray.add(DataAwareStepTypes.MINVI);
				} else {
					stepArray.add(DataAwareStepTypes.MREAL);
				}
			} else if (processStep.equals(ExecutionStep.bottomStep)) {
				labelArray.add("P: -  L: " + logStep.toString());
				stepArray.add(DataAwareStepTypes.L);
			} else {
				if (processStep.equals(logStep)) {
					stepArray.add(DataAwareStepTypes.LMGOOD);
				} else {
					stepArray.add(DataAwareStepTypes.LMNOGOOD);
				}
				labelArray.add("P: " + processStep + "  L: " + logStep);

			}
		}

		return align;
	}

	protected float computeFitness(ReplayState state) {
		float numerator = state.getCost() - lengthWithNoBottomSteps(state.getLogTracePrefix()) * minCost;
		float denumerator = 0;
		for (ExecutionStep s : state.getLogTracePrefix()) {
			if (!s.equals(ExecutionStep.bottomStep))
				denumerator += activityCost.costMoveOnlyInLog(s.getActivity());
		}
		for (ExecutionStep s : state.getProcessTracePrefix()) {
			if (!s.equals(ExecutionStep.bottomStep))
				denumerator += activityCost.costMoveOnlyInProcess(s.getActivity());
		}
		if (denumerator > 0)
			return 1 - numerator / denumerator;
		else
			return 0;

	}

	private static float lengthWithNoBottomSteps(List<ExecutionStep> logTracePrefix) {
		int num = 0;
		for (ExecutionStep s : logTracePrefix)
			if (s.equals(ExecutionStep.bottomStep))
				num++;
		return (logTracePrefix.size() - num);
	}

	/**
	 * Mean and variance are computed according to the Knuth's algorithm at page
	 * 232 of book Donald E. Knuth (1998). The Art of Computer Programming,
	 * volume 2: Seminumerical Algorithms, 3rd edn., Addison-Wesley.
	 */
	protected void updateStatisticsMap(ExecutionStep logStep, ExecutionStep processStep) {
		Set<String> varSet = new HashSet<>(processStep.keySet());
		String actName;
		if (!processStep.equals(ExecutionStep.bottomStep)) {
			actName = processStep.getActivity();
			varSet.addAll(processStep.keySet());
		} else {
			actName = logStep.getActivity();
			varSet.addAll(logStep.keySet());
		}
		float actStat[] = actArray.get(actName);
		if (actStat == null) {
			actStat = new float[4];
			actArray.put(actName, actStat);
		}
		if (logStep.equals(ExecutionStep.bottomStep)) {
			actStat[3]++;
		} else if (processStep.equals(ExecutionStep.bottomStep)) {
			actStat[2]++;
			return; // Don't go on with attributes for a Move In Log
		} else {
			if (logStep.equals(processStep))
				actStat[0]++;
			else
				actStat[1]++;
		}
		for (String varName : varSet) {
			float[] attrStat = attrArray.get(varName);
			if (attrStat == null) {
				attrStat = new float[5];
				attrArray.put(varName, attrStat);
			}

			if (logStep.get(varName) == null || processStep.get(varName) == null) {
				attrStat[1]++;
			} else {
				double diff = Math.abs(org.processmining.models.guards.NumericValueConversion
						.toNumericValue(processStep.get(varName))
						- org.processmining.models.guards.NumericValueConversion.toNumericValue(logStep.get(varName)));
				double delta = diff - attrStat[2];
				attrStat[2] += (delta) / (attrStat[0] - attrStat[1]);
				attrStat[3] += delta * (diff - attrStat[2]);
				if (diff > 0.1)
					attrStat[4]++;
				else
					attrStat[0]++;
			}

		}

	}

	public float getMeanVisitedNode() {
		return meanVisitedNode;
	}

	public float getMeanBuiltTreeSize() {
		return meanBuiltTreeSize;
	}

	public XLog getAlignedLog() {
		return alignedLog;
	}

	/**
	 * @return mapping from variable name in DPN-net to variable name in event log
	 */
	public Map<String, String> getVariableMapping() {
		return variableMapping;
	}

	public Alignment getAlignmentByTraceName(String traceName) {
		return nameToAlignmentMap.get(traceName);
	}

	public Collection<Alignment> getAlignments() {
		return labelStepArray;
	}

	public Iterator<Alignment> iterator() {
		return labelStepArray.iterator();
	}

	public XEventClassifier getClassifier() {
		return classifier;
	}
	
}
