package org.processmining.plugins.DataConformance;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.StringTokenizer;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.DataConformance.framework.Replayable;

import com.google.common.primitives.Doubles;
import com.google.common.primitives.Floats;
import com.google.common.primitives.Ints;
import com.google.common.primitives.Longs;

public class Utility {

	public static String encodeTextAsHTML(String pText) {
		StringTokenizer tokenizer = new StringTokenizer(pText, "&<>\"", true);
		int tokenCount = tokenizer.countTokens();

		/* no encoding's needed */
		if (tokenCount == 1)
			return pText;

		/*
		 * text.length + (tokenCount * 6) gives buffer large enough so no
		 * addition memory would be needed and no costly copy operations would
		 * occur
		 */
		StringBuffer buffer = new StringBuffer(pText.length() + tokenCount * 6);
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			if (token.length() == 1) {
				switch (token.charAt(0)) {
					case '&' :
						buffer.append("&amp;");
						break;
					case '<' :
						buffer.append("&lt;");
						break;
					case '>' :
						buffer.append("&gt;");
						break;
					case '"' :
						buffer.append("&quot;");
						break;
					default :
						buffer.append(token);
				}
			} else {
				buffer.append(token);
			}
		}
		return buffer.toString();
	}

	public static float stringDistance(String first, String second) {
		return (float) Math.abs(org.processmining.models.guards.NumericValueConversion.convertToFloat(first)
				- org.processmining.models.guards.NumericValueConversion.convertToFloat(second));
	}

	public static Object getValue(XAttribute attribute) {
		if (attribute instanceof XAttributeBoolean)
			return ((XAttributeBoolean) attribute).getValue();
		else if (attribute instanceof XAttributeContinuous)
			return ((XAttributeContinuous) attribute).getValue();
		else if (attribute instanceof XAttributeDiscrete)
			return ((XAttributeDiscrete) attribute).getValue();
		else if (attribute instanceof XAttributeLiteral)
			return ((XAttributeLiteral) attribute).getValue();
		else if (attribute instanceof XAttributeTimestamp)
			return ((XAttributeTimestamp) attribute).getValue();
		else
			return null;
	}

	public static Map<String, Float> getMaxDistanceInValue(Replayable model, XLog log,
			Map<String, String> variableMapping) {
		HashMap<String, Float> maxProcess = new HashMap<String, Float>();
		HashMap<String, Float> minProcess = new HashMap<String, Float>();
		HashMap<String, Float> maxLog = new HashMap<String, Float>();
		HashMap<String, Float> minLog = new HashMap<String, Float>();
		for (String var : variableMapping.keySet()) {
			Set<Comparable> values = model.getSetValues(model.getSetForVariable(var));
			float minValueProcess = Float.MAX_VALUE;
			float maxValueProcess = Float.MIN_VALUE;
			for (Comparable value : values) {
				float processValue = (float) org.processmining.models.guards.NumericValueConversion
						.toNumericValue(value);
				if (processValue < minValueProcess) {
					minValueProcess = processValue;
					minProcess.put(var, processValue);
				}
				if (processValue > maxValueProcess) {
					maxValueProcess = processValue;
					maxProcess.put(var, processValue);
				}
			}
		}
		for (String var : variableMapping.values()) {
			maxLog.put(var, Float.MIN_VALUE);
			minLog.put(var, Float.MAX_VALUE);
		}

		for (XTrace trace : log)
			for (XEvent event : trace)
				for (Entry<String, XAttribute> entry : event.getAttributes().entrySet())
					if (maxLog.containsKey(entry.getKey())) {
						float maxValueLog = maxLog.get(entry.getKey());
						float minValueLog = minLog.get(entry.getKey());
						float logValue = (float) org.processmining.models.guards.NumericValueConversion
								.toNumericValue(getValue(entry.getValue()));
						if (logValue < minValueLog)
							minLog.put(entry.getKey(), logValue);
						if (logValue > maxValueLog)
							maxLog.put(entry.getKey(), logValue);
					}

		Map<String, Float> retValue = new HashMap<String, Float>();
		for (Entry<String, String> entry : variableMapping.entrySet()) {
			float minProcessValue = minProcess.get(entry.getKey());
			float maxProcessValue = maxProcess.get(entry.getKey());
			float minLogValue = minLog.get(entry.getValue());
			float maxLogValue = minLog.get(entry.getValue());
			float maxDistance = Math.max(maxLogValue, maxProcessValue) - Math.min(minProcessValue, minLogValue);
			retValue.put(entry.getKey(), maxDistance);
		}
		return (retValue);
	}

	public static Object getValue(XAttribute value, Class<?> varType) {
		Object retValue = getValue(value);
		if (varType == Integer.class) {
			try {
				retValue = Integer.parseInt(retValue.toString());
			} catch (NumberFormatException err) {
				retValue = 0;
			}
		} else if (varType == Float.class) {
			try {
				retValue = Float.parseFloat(retValue.toString());
			} catch (NumberFormatException err) {
				retValue = 0F;
			}
		} else if (varType == Long.class) {
			try {
				retValue = Long.parseLong(retValue.toString());
			} catch (NumberFormatException err) {
				retValue = 0L;
			}
		} else if (varType == Double.class) {
			try {
				retValue = Double.parseDouble(retValue.toString());
			} catch (NumberFormatException err) {
				retValue = 0D;
			}
		} else if (varType == Boolean.class) {
			try {
				retValue = Boolean.parseBoolean(retValue.toString());
			} catch (NumberFormatException err) {
				retValue = false;
			}
		}
		return retValue;
	}

	/**
	 * Checks if the type of value is the 'same' as the expected type. Here the
	 * 'same' means that there there will be no loss in precision or overflow
	 * when casting value to expected type.
	 * 
	 * @param value
	 * @param expectedType
	 * @return false if value is not assignable without precision loss
	 */
	public static boolean isAssignableToExpectedType(Object value, Class<?> expectedType) {
		if (value instanceof Long) {
			boolean isExpected = expectedType == Long.class || expectedType == Double.class;
			if (!isExpected) {
				return ((Long) value).longValue() < Integer.MAX_VALUE && ((Long) value).longValue() > Integer.MIN_VALUE;
			} else {
				return isExpected;
			}
		} else if (value instanceof Integer) {
			return expectedType == Long.class || expectedType == Integer.class;
		} else if (value instanceof Float) {
			return expectedType == Double.class || expectedType == Float.class;
		} else if (value instanceof String) {
			String strVal = (String) value;
			if (expectedType == String.class) {
				return true;
			} else if (expectedType == Long.class) {
				return Longs.tryParse(strVal) != null;
			} else if (expectedType == Integer.class) {
				return Ints.tryParse(strVal) != null;
			} else if (expectedType == Double.class) {
				return Doubles.tryParse(strVal) != null;
			} else if (expectedType == Float.class) {
				return Floats.tryParse(strVal) != null;
			} else {
				//TODO support dates here?
				return false;
			}
		} else {
			return expectedType.isInstance(value);
		}
	}

	/**
	 * Converts the value to the expected type without doing any checks. Throws
	 * {@link IllegalArgumentException} or {@link NumberFormatException} if used
	 * incorrectly.
	 * 
	 * @param value
	 * @param expectedType
	 * @return
	 */
	public static Object convertToExpectedType(Object value, Class<?> expectedType) {
		if (expectedType == Long.class) {
			if (value instanceof Long) {
				return value;
			} else if (value instanceof Integer) {
				return Long.valueOf(((Integer) value).longValue());
			} else if (value instanceof String) {
				return Long.parseLong((String) value);
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == Integer.class) {
			if (value instanceof Integer) {
				return value;
			} else if (value instanceof Long) {
				return Integer.valueOf(((Long) value).intValue());
			} else if (value instanceof String) {
				return Integer.parseInt((String) value);
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == Float.class) {
			if (value instanceof Float) {
				return value;
			} else if (value instanceof String) {
				return Float.parseFloat((String) value);
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == Double.class) {
			if (value instanceof Long) {
				return ((Long) value).doubleValue();
			} else if (value instanceof Float) {
				return Double.valueOf(((Float) value).doubleValue());
			} else if (value instanceof Double) {
				return value;
			} else if (value instanceof String) {
				return Double.parseDouble((String) value);
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == String.class) {
			if (value instanceof String) {
				return value;
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == Boolean.class) {
			if (value instanceof Boolean) {
				return value;
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		} else if (expectedType == Date.class) {
			if (value instanceof Date) {
				return value;
			} else {
				throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
			}
		}
		throw new IllegalArgumentException("Not convertable " + value + " to " + expectedType.getSimpleName());
	}

}
