package org.processmining.plugins.DataConformance.framework;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

public class VariableMatchCosts implements WriteOperationCost {

	private static class Pair {
		String activity;
		String variable;

		public Pair(String act, String var) {
			activity = act;
			variable = var;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((activity == null) ? 0 : activity.hashCode());
			result = prime * result + ((variable == null) ? 0 : variable.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Pair other = (Pair) obj;
			if (activity == null) {
				if (other.activity != null)
					return false;
			} else if (!activity.equals(other.activity))
				return false;
			if (variable == null) {
				if (other.variable != null)
					return false;
			} else if (!variable.equals(other.variable))
				return false;
			return true;
		}

		public String toString() {
			return "(" + activity + "," + variable + ")";
		}
	}

	/**
	 * The list of {@link VariableMatchCost} that this instance was built with
	 */
	private final List<VariableMatchCost> variableCosts;

	// Lookup maps
	private final Map<Pair, Float> costFaultyValueMap;
	private final Map<Pair, Float> costNotWritingMap;
	private final Set<Pair> finalVariablesSet;

	private float lowestCost = Float.MAX_VALUE;
	private float highestCost = Float.MIN_VALUE;

	public static VariableMatchCosts NOCOST = new VariableMatchCosts();

	private VariableMatchCosts() {
		highestCost = Float.MAX_VALUE;
		this.costFaultyValueMap = ImmutableMap.of();
		this.costNotWritingMap = ImmutableMap.of();
		this.finalVariablesSet = ImmutableSet.of();
		this.variableCosts = ImmutableList.of();
	}

	/**
	 * Initialize the {@link VariableMatchCosts} instance with the specified
	 * defaultCostFaulty and defaultCostNotWriting values for all activities and
	 * attributes.
	 * 
	 * @param defaultCostFaulty
	 * @param defaultCostNotWriting
	 * @param activitySet
	 * @param variableSet
	 */
	public VariableMatchCosts(int defaultCostFaulty, int defaultCostNotWriting, Set<ReplayableActivity> activitySet,
			Set<String> variableSet) {
		this(ImmutableList.of(new VariableMatchCost(defaultCostFaulty, defaultCostNotWriting)), activitySet,
				variableSet);
	}

	/**
	 * Initialize the {@link VariableMatchCosts} instance with the specified
	 * list of {@link VariableMatchCost}. Please note: The order of the list is
	 * important. Later elements may overwrite over earlier ones.
	 * 
	 * @param variableCosts
	 * @param activitySet
	 * @param variableSet
	 */
	public VariableMatchCosts(List<VariableMatchCost> variableCosts, Set<? extends ReplayableActivity> activitySet,
			Set<String> variableSet) {
		this.variableCosts = variableCosts;
		Map<Pair, Float> costFaultyValueMap = new HashMap<Pair, Float>();
		Map<Pair, Float> costNotWritingMap = new HashMap<Pair, Float>();
		Set<Pair> finalVariablesSet = new HashSet<Pair>();
		for (ReplayableActivity node : activitySet)
			for (String variable : variableSet)
				for (VariableMatchCost matchCost : variableCosts) {
					String varMatch = matchCost.getVariable();
					String actMatch = matchCost.getActivity();
					if (varMatch == null && actMatch == null || varMatch == null && actMatch.equals(node.getLabel())
							|| variable.equals(varMatch) && actMatch == null
							|| variable.equals(varMatch) && node.getLabel().equals(actMatch)) {
						Pair activityVarPair = new Pair(node.getLabel(), variable);
						costFaultyValueMap.put(activityVarPair, matchCost.getCostFaultyValue());
						costNotWritingMap.put(activityVarPair, matchCost.getCostNotWriting());
						if (matchCost.isFinal()) {
							finalVariablesSet.add(activityVarPair);
						}

						if (Math.min(matchCost.getCostFaultyValue(), matchCost.getCostNotWriting()) < lowestCost)
							lowestCost = Math.min(matchCost.getCostFaultyValue(), matchCost.getCostNotWriting());
						if (Math.max(matchCost.getCostFaultyValue(), matchCost.getCostNotWriting()) > highestCost)
							highestCost = Math.max(matchCost.getCostFaultyValue(), matchCost.getCostNotWriting());

						break;
					}
				}
		this.costFaultyValueMap = ImmutableMap.copyOf(costFaultyValueMap);
		this.costNotWritingMap = ImmutableMap.copyOf(costNotWritingMap);
		this.finalVariablesSet = ImmutableSet.copyOf(finalVariablesSet);
	}

	public float lowestCost() {
		return lowestCost;
	}

	public float highestCost() {
		return highestCost;
	}

	public float costFaultyValue(String activity, String var) {
		Float cost = costFaultyValueMap.get(new Pair(activity, var));
		if (cost == null)
			return (float) ((lowestCost + highestCost) / 2.0); //TODO WTF, is this intended behavior?
		assert (cost != null);
		return cost;
	}

	public float notWritingCost(String activity, String var) {
		Float cost = costNotWritingMap.get(new Pair(activity, var));
		assert (cost != null);
		return cost;
	}

	public boolean isFinal(String activity, String var) {
		return finalVariablesSet.contains(new Pair(activity, var));
	}

	public List<VariableMatchCost> getVariableCosts() {
		return variableCosts;
	}

}