package org.processmining.plugins.DataConformance.visualization.comparison;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.DefaultTrace;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.DataConformance.visualization.DataAwareStepTypes;
import org.processmining.plugins.utils.sequencedistance.GenericLevenshteinDistance;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance.DistanceResult;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance.Equivalence;

import com.google.common.base.Function;
import com.google.common.collect.Lists;

public class AlignmentComparisonPlugin {

	private static class AlignmentStep {

		public AlignmentStep(DataAwareStepTypes type, ExecutionStep processStep, ExecutionStep logStep) {
			this.type = type;
			this.processStep = processStep;
			this.logStep = logStep;
			this.label = processStep.getActivity() != null ? processStep.getActivity() : logStep.getActivity();
		}

		DataAwareStepTypes type;
		ExecutionStep processStep;
		ExecutionStep logStep;
		String label;

	}

	private final static class AlignmentEntryImpl implements AlignmentEntry {

		private List<AlignmentStep> alignmentA;
		private List<AlignmentStep> alignmentB;
		private int maxLength;
		private int minLength;
		private String name;
		private Trace<Event> combinedAlignment;
		private int distance;
		private DistanceResult<AlignmentStep> levenshteinResult;
		private float fitnessA;
		private float fitnessB;

		private AlignmentEntryImpl(String name) {
			super();
			this.name = name;
		}

		public String getName() {
			return name;
		}

		public int getMaxLength() {
			return maxLength;
		}

		public int getMinLength() {
			return minLength;
		}

		public void setAlignmentA(Alignment a) {
			alignmentA = convert(a);
			maxLength = Math.max(maxLength, a.getStepTypes().size());
			minLength = Math.min(minLength, a.getStepTypes().size());
			fitnessA = a.getFitness();
		}

		public void setAlignmentB(Alignment b) {
			alignmentB = convert(b);
			maxLength = Math.max(maxLength, b.getStepTypes().size());
			minLength = Math.min(minLength, b.getStepTypes().size());
			fitnessB = b.getFitness();
		}

		private List<AlignmentStep> convert(Alignment a) {
			int size = a.getStepTypes().size();
			List<AlignmentStep> steps = new ArrayList<>(size);
			for (int i = 0; i < size; i++) {
				steps.add(new AlignmentStep(a.getStepTypes().get(i), a.getLogTrace().get(i), a.getProcessTrace().get(i)));
			}
			return steps;
		}

		public List<AlignmentStep> getAlignmentB() {
			return alignmentB;
		}

		public List<AlignmentStep> getAlignmentA() {
			return alignmentA;
		}

		public void buildCombinedTrace() {
			if (alignmentA != null && alignmentB != null) {

				SequenceDistance<AlignmentStep> levenshteinDistance = new GenericLevenshteinDistance<>();
				levenshteinResult = levenshteinDistance.computeAlignment(alignmentA, alignmentB,
						new Equivalence<AlignmentComparisonPlugin.AlignmentStep>() {

							public boolean equals(AlignmentStep a, AlignmentStep b) {
								//TODO what about two transitions with the same name? 
								return a.label.equals(b.label);
							}
						});

				combinedAlignment = new ProMTraceView.AbstractTrace<Event>() {

					public String getName() {
						return "Trace: " + name + " | Fitness A: " + NumberFormat.getPercentInstance().format(fitnessA);
					}

					public String getInfo() {
						return "Distance: " + String.valueOf(getDistance()) + " | Fitness B: "
								+ NumberFormat.getPercentInstance().format(fitnessB);
					}

					private int getDistance() {
						return levenshteinResult.getDistance();
					}

					protected List<Event> delegate() {
						return Lists.transform(levenshteinResult.getAlignment(),
								new Function<DistanceResult.Entry<AlignmentStep>, Event>() {

									public Event apply(DistanceResult.Entry<AlignmentStep> entry) {
										AlignmentStep a = entry.getA();
										AlignmentStep b = entry.getB();
										if (a == null) {
											return new SplittedDataAlignmentEvent(null, null, null, b.processStep,
													b.logStep, b.type);
										} else if (b == null) {
											return new SplittedDataAlignmentEvent(a.processStep, a.logStep, a.type,
													null, null, null);
										} else if (a.type != b.type) {
											return new SplittedDataAlignmentEvent(a.processStep, a.logStep, a.type,
													b.processStep, b.logStep, b.type);
										} else {
											return new AlignmentComparisonEvent(a.processStep, a.logStep, a.type);
										}
									}

								});
					}

				};
				distance = levenshteinResult.getDistance();

			} else {
				DefaultTrace<Event> trace = new ProMTraceView.DefaultTrace<>(name, "Distance: \u221E", maxLength);
				if (alignmentA == null) {
					for (AlignmentStep e : alignmentB) {
						trace.add(new SplittedDataAlignmentEvent(null, null, null, e.processStep, e.logStep, e.type));
					}
				} else {
					for (AlignmentStep e : alignmentA) {
						trace.add(new SplittedDataAlignmentEvent(e.processStep, e.logStep, e.type, null, null, null));
					}
				}
				combinedAlignment = trace;
				distance = Integer.MAX_VALUE;
			}
		}

		public Trace<? extends Event> getCombinedTrace() {
			return combinedAlignment;
		}

		public int getDistance() {
			return distance;
		}

		public DistanceResult<AlignmentStep> getLevenshteinResult() {
			return levenshteinResult;
		}

	}

	public interface AlignmentEntry {

		String getName();

		DistanceResult<AlignmentStep> getLevenshteinResult();

		void setAlignmentA(Alignment trace);

		void setAlignmentB(Alignment trace);

		List<AlignmentStep> getAlignmentA();

		List<AlignmentStep> getAlignmentB();

		int getMinLength();

		int getMaxLength();

		int getDistance();

		void buildCombinedTrace();

		Trace<? extends Event> getCombinedTrace();

	}

	public interface AlignmentComparisonResult {

		Collection<AlignmentEntry> getTraceEntries();

	}

	@Plugin(name = "Compare Two Alignments", level = PluginLevel.NightlyBuild, returnLabels = { "Alignment Comparison" }, returnTypes = { AlignmentComparisonResult.class }, userAccessible = true, parameterLabels = {
			"Alignment A", "Alignment B" }, mostSignificantResult = 1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "DataAwareReplayer")
	public AlignmentComparisonResult compareLogs(PluginContext context, final ResultReplay resultA,
			final ResultReplay resultB) throws Exception {
		context.getProgress().setCaption("Comparing logs ...");
		final Collection<AlignmentEntry> traceEntries = buildTraceEntries(resultA, resultB);
		return new AlignmentComparisonResult() {

			public Collection<AlignmentEntry> getTraceEntries() {
				return traceEntries;
			}
		};
	}

	private Collection<AlignmentEntry> buildTraceEntries(ResultReplay resultA, ResultReplay resultB) throws Exception {

		Map<String, AlignmentEntry> entriesByName = new HashMap<>();

		for (Alignment alignment : resultA.labelStepArray) {
			String name = alignment.getTraceName();
			if (name != null) {
				AlignmentEntry alignmentEntry = new AlignmentEntryImpl(name);
				entriesByName.put(name, alignmentEntry);
				alignmentEntry.setAlignmentA(alignment);
			} else {
				throw new Exception("Trace is missing name " + alignment);
			}
		}

		for (Alignment alignment : resultB.labelStepArray) {
			String name = alignment.getTraceName();
			if (name != null) {
				AlignmentEntry traceEntry = entriesByName.get(name);
				if (traceEntry == null) {
					traceEntry = new AlignmentEntryImpl(name);
				}
				traceEntry.setAlignmentB(alignment);
			} else {
				throw new Exception("Trace is missing name " + alignment);
			}
		}

		for (AlignmentEntry e : entriesByName.values()) {
			e.buildCombinedTrace();
		}

		return entriesByName.values();
	}

}
