/*
 * Mainly copied from nl.tue.astar.util.AbstractFastLookupPriorityQueue
 */
package org.processmining.plugins.balancedconformance.controlflow.override.queues.noduplicate;

import java.util.Arrays;
import java.util.Collection;
import java.util.Queue;

import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;

import it.unimi.dsi.fastutil.Hash;
import it.unimi.dsi.fastutil.Hash.Strategy;
import it.unimi.dsi.fastutil.objects.Object2IntOpenCustomHashMap;
import nl.tue.astar.Record;
import nl.tue.astar.Tail;
import nl.tue.astar.util.AbstractFastLookupPriorityQueue;
import nl.tue.astar.util.FastLookupPriorityQueue;

/**
 * Implementation of a {@link FastLookupPriorityQueue} that allows to store
 * Objects instead of long values in the queue. Most code taken from
 * {@link AbstractFastLookupPriorityQueue}.
 * 
 * @author F. Mannhardt
 * 
 */
public abstract class DataFastLookupPriorityQueue implements FastLookupPriorityQueue {

	protected final Object2IntOpenCustomHashMap<DRecord> locationMap;
	protected static final int NEV = -1;

	/**
	 * Priority queue represented as a balanced binary heap: the two children of
	 * queue[n] are queue[2*n+1] and queue[2*(n+1)]. The priority queue is
	 * ordered by the record's natural ordering: For each node n in the heap and
	 * each descendant d of n, n <= d. The element with the best value is in
	 * queue[0], assuming the queue is nonempty.
	 */
	protected DRecord[] queue;

	/**
	 * The number of elements in the priority queue.
	 */
	protected int size = 0;

	/**
	 * The maximum total cost for any record in this queue. If the cost of a
	 * record which is added is higher that this value, it is not added
	 */
	protected int maxCost;

	public DataFastLookupPriorityQueue(int initialCapacity, Strategy<DRecord> strategy,
			AbstractPDelegate<? extends Tail> delegate) {
		this(initialCapacity, strategy, delegate, Integer.MAX_VALUE);
	}

	public DataFastLookupPriorityQueue(int initialCapacity, Strategy<DRecord> strategy,
			AbstractPDelegate<? extends Tail> delegate, int maxCost) {
		this.maxCost = maxCost;
		this.locationMap = new Object2IntOpenCustomHashMap<>(initialCapacity, Hash.FAST_LOAD_FACTOR, strategy);
		this.locationMap.defaultReturnValue(NEV);
		this.queue = new DRecord[initialCapacity];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#isEmpty()
	 */
	public boolean isEmpty() {
		return size() == 0;
	}

	public int hashCode() {
		return locationMap.hashCode();
	}

	public boolean equals(Object o) {
		return (o instanceof DataFastLookupPriorityQueue)
				&& ((DataFastLookupPriorityQueue) o).locationMap.equals(locationMap);
	}

	public void setMaxCost(int maxCost) {
		this.maxCost = maxCost;
	}

	public int getMaxCost() {
		return this.maxCost;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * nl.tue.astar.util.FastLookupPriorityQueue#contains(nl.tue.astar.Record)
	 */
	public DRecord contains(Record newRec) {
		int location = locationMap.getInt(newRec);
		if (location != NEV) {
			return peek(location);
		} else {
			return null;
		}
	}

	public boolean checkInv() {
		return checkInv(0);
	}

	/**
	 * Increases the capacity of the array.
	 * 
	 * @param minCapacity
	 *            the desired minimum capacity
	 */
	protected void grow(int minCapacity) {
		if (minCapacity < 0) // overflow
			throw new OutOfMemoryError();
		int oldCapacity = queue.length;
		// Double size if small; else grow by 50%
		int newCapacity = ((oldCapacity < 64) ? ((oldCapacity + 1) * 2) : ((oldCapacity / 2) * 3));
		if (newCapacity < 0) // overflow
			newCapacity = Integer.MAX_VALUE;
		if (newCapacity < minCapacity)
			newCapacity = minCapacity;
		queue = Arrays.copyOf(queue, newCapacity);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#peek()
	 */
	public Record peek() {
		if (size == 0)
			return null;
		return queue[0];
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#size()
	 */
	public int size() {
		return size;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see nl.tue.astar.util.FastLookupPriorityQueue#poll()
	 */
	public Record poll() {
		if (size == 0)
			return null;
		int s = --size;
		DRecord result = queue[0];
		DRecord x = queue[s];
		queue[s] = null;
		locationMap.remove(result);

		if (s != 0)
			siftDown(0, x);
		return result;
	}

	private DRecord peek(int location) {
		return queue[location];
	}

	public String toString() {
		return Arrays.toString(queue);
	}

	private void updateToBetter(int index, Record newRecord) {
		assert index >= 0 && index < size;
		assert index >= 0 : "k < 0 for record " + newRecord.toString() + " queue " + toString();
		locationMap.put(((DRecord) newRecord), index);
		siftUp(index, (DRecord) newRecord);
	}

	/**
	 * Inserts the specified element into this priority queue.
	 * 
	 * @return {@code true} (as specified by {@link Collection#add})
	 * @throws ClassCastException
	 *             if the specified element cannot be compared with elements
	 *             currently in this priority queue according to the priority
	 *             queue's ordering
	 * @throws NullPointerException
	 *             if the specified element is null
	 */
	public boolean add(Record newE) {
		if (newE.getTotalCost() > maxCost) {
			return false;
		}
		// check if overwrite is necessary, i.e. only add if the object does not
		// exist yet,
		// or exists, but with higher costs.
		int location = locationMap.getInt(newE);
		if (location == NEV) {
			// new element, add to queue and return
			offer((DRecord) newE);
			return true;
		}

		if (isBetter((DRecord) newE, peek(location))) {
			// update to better, if newE better then peek(location)
			updateToBetter(location, newE);
			return true;
		}
		return false;
	}

	protected abstract boolean isBetter(DRecord r1, DRecord r2);

	/**
	 * Inserts the specified element into this priority queue.
	 * 
	 * @return {@code true} (as specified by {@link Queue#offer})
	 * @throws ClassCastException
	 *             if the specified element cannot be compared with elements
	 *             currently in this priority queue according to the priority
	 *             queue's ordering
	 * @throws NullPointerException
	 *             if the specified element is null
	 */
	private void offer(DRecord e) {
		if (e == null)
			throw new NullPointerException();
		int i = size;
		if (i >= queue.length)
			grow(i + 1);
		size = i + 1;
		if (i == 0) {
			queue[0] = e;
			locationMap.put(e, 0);
		} else
			siftUp(i, e);
	}

	/**
	 * Inserts item x at position k, maintaining heap invariant by promoting x
	 * up the tree until it is greater than or equal to its parent, or is the
	 * root.
	 * 
	 * @param k
	 * @param x
	 *            the item to insert
	 */
	private void siftUp(int k, DRecord x) {
		while (k > 0) {
			int parent = (k - 1) >>> 1;
			DRecord e = queue[parent];
			if (!isBetter(x, e)) {
				break;
			}
			queue[k] = e;
			assert k >= 0 : "k < 0 for record " + x.toString() + " queue " + toString();
			locationMap.put(e, k);
			k = parent;
		}
		queue[k] = x;
		assert k >= 0 : "k < 0 for record " + x.toString() + " queue " + toString();
		locationMap.put(x, k);
	}

	/**
	 * Inserts item x at position k, maintaining heap invariant by demoting x
	 * down the tree repeatedly until it is less than or equal to its children
	 * or is a leaf.
	 * 
	 * @param k
	 *            the position to fill
	 * @param x
	 *            the item to insert
	 */
	private void siftDown(int k, DRecord x) {
		int half = size >>> 1;
		while (k < half) {
			int child = (k << 1) + 1;
			DRecord c = queue[child];
			int right = child + 1;
			if (right < size && isBetter(queue[right], c))
				c = queue[child = right];

			if (!isBetter(c, x))
				break;

			queue[k] = c;
			//assert locationMap.get(c) == child;
			// i.e. child + k -child == k,
			// hence we use adjustValue instead of put here.
			locationMap.addTo(c, k - child);
			assert locationMap.getInt(c) >= 0 : "k < 0 for record " + x.toString() + " queue " + toString();
			k = child;
		}
		queue[k] = x;
		assert k >= 0 : "k < 0 for record " + x.toString() + " queue " + toString();
		locationMap.put(x, k);
	}

	protected boolean checkInv(int loc) {
		DRecord n = queue[loc];
		DRecord c1 = null;
		DRecord c2 = null;
		if (2 * loc + 1 < queue.length)
			c1 = queue[2 * loc + 1];

		if (2 * (loc + 1) < queue.length)
			c2 = queue[2 * (loc + 1)];

		return (c1 == null ? true : !isBetter(c1, n) && checkInv(2 * loc + 1))
				&& (c2 == null ? true : !isBetter(c2, n) && checkInv(2 * (loc + 1)));
	}

}
