/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow;

import org.processmining.plugins.DataConformance.DataAlignment.GenericTrace;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentResult.Step;
import org.processmining.plugins.balancedconformance.controlflow.override.DRecord;
import org.processmining.plugins.balancedconformance.controlflow.override.DataAwareThread.PartialResultFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateStore;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.lp.gurobi.DataAlignmentSolverFactoryGurobi;
import org.processmining.plugins.balancedconformance.dataflow.lp.lpsolve.DataAlignmentSolverFactoryLpSolve;
import org.processmining.plugins.balancedconformance.mapping.LogMapping;
import org.processmining.plugins.balancedconformance.mapping.TraceMapping;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;
import org.processmining.plugins.balancedconformance.result.DataAlignedTraceImpl;
import org.processmining.plugins.petrinet.replayresult.StepTypes;

import com.google.common.collect.ImmutableList;

/**
 * Interface for building a data alignment given a trace and a control flow
 * alignment
 * 
 * @author F. Mannhardt
 * 
 */
public interface DataAlignmentAdapter extends AutoCloseable {

	public static enum ILPSolver {
		ILP_LPSOLVE("LpSolve (Bundled)"), 
		ILP_GUROBI("Gurobi (Experimental: License & Gurobi installation required)");

		private String desc;

		private ILPSolver(String desc) {
			this.desc = desc;
		}

		@Override
		public String toString() {
			return desc;
		}

	}

	public static class LpSolveDataAlignmentAdapter extends DataAlignmentAdapterMILPImpl {

		public LpSolveDataAlignmentAdapter(LogMapping logMapping, DataStateFactory dataStateFactory,
				int concurrentThreads, double aStarDelta, boolean includeVirtualVariablesInTrace, boolean usePartialMILPProblems) {
			super(logMapping, new DataAlignmentSolverFactoryLpSolve(concurrentThreads), dataStateFactory, aStarDelta,
					includeVirtualVariablesInTrace, usePartialMILPProblems);
		}

	}

	public static class LpSolveWithoutPartialDataAlignmentAdapter extends DataAlignmentAdapterWithoutPartialImpl {

		public LpSolveWithoutPartialDataAlignmentAdapter(LogMapping logMapping, DataStateFactory dataStateFactory,
				int concurrentThreads, double aStarDelta, boolean includeVirtualVariablesInTrace) {
			super(logMapping, new DataAlignmentSolverFactoryLpSolve(concurrentThreads), dataStateFactory, aStarDelta,
					includeVirtualVariablesInTrace);
		}

	}

	public static class NoDataAlignmentAdapter implements DataAlignmentAdapter {

		private static final PartialDataAlignedTrace EMPTY_DATA_ALIGNMENT = new PartialDataAlignedTrace(0,
				DataStateStore.EMPTY_IMMUTABLE_STATE);

		private DataStateFactory dataStateFactory;

		public NoDataAlignmentAdapter(LogMapping logMapping, DataStateFactory dataStateFactory) {
			super();
			this.dataStateFactory = dataStateFactory;
		}

		public DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
				ControlFlowAlignmentResult controlFlowAlignment) throws DataAlignmentException {
			ImmutableList<Step> stepList = ImmutableList.copyOf(controlFlowAlignment.getSteps());

			final GenericTrace logTrace = new GenericTrace(stepList.size(), traceMapping.getTraceName());
			final GenericTrace processTrace = new GenericTrace(stepList.size(), traceMapping.getTraceName());

			for (Step step : stepList) {

				final String activityName = step.getType() == StepTypes.L
						? traceMapping.getLogMapping().getEventClassifier().getClassIdentity(step.getLogNode())
						// workaround for synchronized name lookup in Transition.getLabel()
						: traceMapping.getLogMapping().getTransitionToName().get(step.getModelNode());

				ExecutionStep logStep = ExecutionStep.bottomStep;
				ExecutionStep processStep = ExecutionStep.bottomStep;
				switch (step.getType()) {
					case L :
						logStep = new ExecutionStep(activityName, step.getLogNode());
						break;
					case LMGOOD :
					case LMNOGOOD :
						logStep = new ExecutionStep(activityName, step.getLogNode());
						//$FALL-THROUGH$ // log step already initialized!
					case MREAL :
					case MINVI :
						processStep = new ExecutionStep(activityName, step.getModelNode());
						if (step.getType() == StepTypes.MINVI) {
							processStep.setInvisible(true);
						}
						break;
					default :
						throw new IllegalArgumentException("Unkown step type " + step.getType());

				}

				logTrace.add(logStep);
				processTrace.add(processStep);

			}

			return new DataAlignedTraceImpl(logTrace, processTrace, controlFlowAlignment.getControlFlowCost(), 0);
		}

		public PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
				PartialResultFactory partialResultBuilder, DRecord currentRecord) throws DataAlignmentException {
			return EMPTY_DATA_ALIGNMENT;
		}

		public DataStateFactory getDataStateFactory() {
			return dataStateFactory;
		}

		public void close() {
		}

	}

	public static class GurobiDataAlignmentAdapter extends DataAlignmentAdapterMILPImpl {

		public GurobiDataAlignmentAdapter(LogMapping logMapping, DataStateFactory dataStateFactory, double aStarDelta,
				boolean includeVirtualVariablesInTrace, boolean usePartialMILPProblems) {
			super(logMapping, new DataAlignmentSolverFactoryGurobi(), dataStateFactory, aStarDelta,
					includeVirtualVariablesInTrace, usePartialMILPProblems);
		}

	}

	DataAlignedTrace retrieveFullDataAlignment(TraceMapping traceMapping,
			ControlFlowAlignmentResult controlFlowAlignment) throws DataAlignmentException;

	PartialDataAlignedTrace retrievePartialDataAlignment(TraceMapping traceMapping,
			PartialResultFactory partialResultBuilder, DRecord currentRecord) throws DataAlignmentException;

	DataStateFactory getDataStateFactory();

	@Override
	void close();

}
