/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.dataflow.datastate;

import java.util.Date;

import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;

import com.google.common.math.DoubleMath;

public final class DataStateHelper {

	private DataStateHelper() {
		super();
	}

	public static Object getObjectFromDataState(DataState state, String varName, Class<?> varType,
			StringDiscretizer discretizer) {
		if (varType == String.class) {
			long longVal = state.getLong(varName);
			assert longVal <= Integer.MAX_VALUE && longVal >= Integer.MIN_VALUE : "Integer variable out of bounds";
			return unDiscretize(varName, (int) longVal, discretizer);
		}
		if (varType == Long.class)
			return state.getLong(varName);
		if (varType == Boolean.class) {
			long longVal = state.getLong(varName);
			assert longVal == 0 && longVal == 1 : "Boolean variable out of bounds";
			return longVal == 1 ? true : false;
		}
		if (varType == Double.class)
			return state.getDouble(varName);
		if (varType == Date.class)
			return new Date(state.getLong(varName));
		if (varType == Integer.class) {
			long longVal = state.getLong(varName);
			assert longVal <= Integer.MAX_VALUE && longVal >= Integer.MIN_VALUE : "Integer variable out of bounds";
			return Integer.valueOf((int) longVal);
		}
		if (varType == Float.class)
			return Float.valueOf((float) state.getDouble(varName));
		throw new IllegalArgumentException("Unkown variable type "
				+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
	}

	public static Object tryGetPrimeOrNormalObjectFromDataState(DataState state, String varName, Class<?> varType,
			StringDiscretizer discretizer) {
		if (varType == String.class) {
			Long longVal = state.tryGetPrimeOrNormalLong(varName);
			if (longVal != null) {
				return unDiscretize(varName, longVal.intValue(), discretizer);
			} else {
				return null;
			}
		}
		if (varType == Long.class) {
			Long longVal = state.tryGetPrimeOrNormalLong(varName);
			if (longVal != null) {
				return longVal;
			} else {
				return null;
			}
		}
		if (varType == Boolean.class) {
			Long longVal = state.tryGetPrimeOrNormalLong(varName);
			if (longVal != null) {
				return longVal == 1 ? true : false;
			} else {
				return null;
			}
		}
		if (varType == Double.class) {
			Double doubleValue = state.tryGetPrimeOrNormalDouble(varName);
			if (doubleValue != null) {
				return doubleValue;
			} else {
				return null;
			}
		}
		if (varType == Date.class) {
			Long longVal = state.tryGetPrimeOrNormalLong(varName);
			if (longVal != null) {
				return new Date(state.getLong(varName));
			} else {
				return null;
			}
		}
		if (varType == Integer.class) {
			Long longVal = state.tryGetPrimeOrNormalLong(varName);
			if (longVal != null) {
				return Integer.valueOf(longVal.intValue());
			} else {
				return null;
			}
		}
		if (varType == Float.class) {
			Double doubleValue = state.tryGetPrimeOrNormalDouble(varName);
			if (doubleValue != null) {
				return Float.valueOf(doubleValue.floatValue());
			} else {
				return null;
			}
		}
		throw new IllegalArgumentException("Unkown variable type "
				+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
	}

	public static Object tryGetObjectFromDataState(DataState state, String varName, Class<?> varType,
			StringDiscretizer discretizer) {
		if (varType == String.class) {
			Long longVal = state.tryGetLong(varName);
			if (longVal != null) {
				return unDiscretize(varName, longVal.intValue(), discretizer);
			} else {
				return null;
			}
		}
		if (varType == Long.class) {
			Long longVal = state.tryGetLong(varName);
			if (longVal != null) {
				return longVal;
			} else {
				return null;
			}
		}
		if (varType == Boolean.class) {
			Long longVal = state.tryGetLong(varName);
			if (longVal != null) {
				return longVal == 1 ? true : false;
			} else {
				return null;
			}
		}
		if (varType == Double.class) {
			Double doubleValue = state.tryGetDouble(varName);
			if (doubleValue != null) {
				return doubleValue;
			} else {
				return null;
			}
		}
		if (varType == Date.class) {
			Long longVal = state.tryGetLong(varName);
			if (longVal != null) {
				return new Date(state.getLong(varName));
			} else {
				return null;
			}
		}
		if (varType == Integer.class) {
			Long longVal = state.tryGetLong(varName);
			if (longVal != null) {
				return Integer.valueOf(longVal.intValue());
			} else {
				return null;
			}
		}
		if (varType == Float.class) {
			Double doubleValue = state.tryGetDouble(varName);
			if (doubleValue != null) {
				return Float.valueOf(doubleValue.floatValue());
			} else {
				return null;
			}
		}
		throw new IllegalArgumentException("Unkown variable type "
				+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
	}

	public static Object getDefaultValue(Class<?> varType) {
		if (varType == String.class)
			return StringDiscretizer.ANY_VALUE;
		if (varType == Integer.class)
			return 0;
		if (varType == Long.class)
			return 0;
		if (varType == Boolean.class)
			return false;
		if (varType == Double.class)
			return 0.0d;
		if (varType == Float.class)
			return 0.0f;
		if (varType == Date.class)
			return new Date(0l);
		throw new IllegalArgumentException("Unkown variable type "
				+ (varType == null ? " null " : varType.getSimpleName()));
	}

	public static void storeValueFromIlp(DataState state, String varName, Number varValue, Class<?> varType) {
		if (varType == String.class) {
			// We can just put the INT value as this is discretized already
			state.putLong(varName, varValue.intValue());
		} else if (varType == Long.class || varType == Integer.class) {
			state.putLong(varName, varValue.longValue());
		} else if (varType == Double.class || varType == Float.class) {
			state.putDouble(varName, varValue.doubleValue());
		} else if (varType == Boolean.class) {
			state.putLong(varName, varValue.intValue());
		} else if (varType == Date.class) {
			state.putLong(varName, varValue.longValue());
		} else {
			throw new IllegalArgumentException("Unkown variable type "
					+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
		}
	}

	public static void storeValue(DataState state, String varName, Object varValue, Class<?> varType,
			StringDiscretizer discretizer) {
		if (varType == String.class) {
			state.putLong(varName, discretize(varName, (String) varValue, discretizer));
		} else if (varType == Long.class || varType == Integer.class) {
			state.putLong(varName, ((Number) varValue).longValue());
		} else if (varType == Boolean.class) {
			state.putLong(varName, ((Boolean) varValue).booleanValue() ? 1 : 0);
		} else if (varType == Double.class || varType == Float.class) {
			state.putDouble(varName, ((Number) varValue).doubleValue());
		} else if (varType == Date.class) {
			state.putLong(varName, ((Date) varValue).getTime());
		} else {
			throw new IllegalArgumentException("Unkown variable type "
					+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
		}
	}

	public static Object convertFromIlpValue(String varName, double value, Class<?> varType,
			StringDiscretizer discretizer) {
		if (varType == String.class) {
			return discretizer.convertToString((int) Math.round(value));
		}
		if (varType == Boolean.class) {
			assert value == 0 || value == 1 : "Boolean variable out of bounds";
			return (value == 0 ? false : true);
		}
		if (varType == Long.class)
			return ((long) Math.round(value));
		if (varType == Double.class)
			return value;
		if (varType == Date.class) {
			assert DoubleMath.isMathematicalInteger(value) : "Date should be an integer!";
			return new Date((long) (value));
		}
		if (varType == Integer.class)
			return ((int) Math.round(value));
		if (varType == Float.class)
			return (float) value;
		throw new IllegalArgumentException("Unkown variable type "
				+ (varType == null ? " null " : varType.getSimpleName()) + " for variable " + varName);
	}

	public static double convertToIlpValue(String varName, Object varValue, StringDiscretizer discretizer) {
		if (varValue instanceof String)
			return discretizer.convertToInt((String) varValue);
		if (varValue instanceof Boolean)
			return ((Boolean) varValue).booleanValue() ? 1 : 0;
		if (varValue instanceof Number)
			return ((Number) varValue).doubleValue();
		if (varValue instanceof Date)
			return ((Date) varValue).getTime();
		throw new IllegalArgumentException("Unkown variable type "
				+ (varValue == null ? " null " : varValue.getClass().getSimpleName()) + " for variable " + varName);
	}

	public static int discretize(String varName, String varValue, StringDiscretizer discretizer) {
		return discretizer.convertToInt(varValue);
	}

	public static String unDiscretize(String varName, int varValue, StringDiscretizer discretizer) {
		return discretizer.convertToString(varValue);
	}

	public static String stripPrime(String varName) {
		return varName.substring(0, varName.length() - 1);
	}

	public static boolean isPrimeVar(String varName) {
		return varName.charAt(varName.length() - 1) == '\'';
	}

	public static String addPrime(String varName) {
		return varName.concat("'");
	}

	public static String stripPrimeIfNeeded(String varName) {
		if (isPrimeVar(varName)) {
			return stripPrime(varName);
		} else {
			return varName;
		}
	}

}