package org.processmining.plugins.balancedconformance.dataflow.lp;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.TimeUnit;

import com.google.common.base.Stopwatch;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

import net.sf.javailp.Constraint;
import net.sf.javailp.Linear;
import net.sf.javailp.Operator;
import net.sf.javailp.OptType;
import net.sf.javailp.Problem;
import net.sf.javailp.Result;
import net.sf.javailp.Solver;
import net.sf.javailp.VarType;

public final class ILPProblemImpl implements ILPProblem {

	private static final String CONSTRAINT_NAME = "";

	private final static class DebugableProblem extends Problem {

		public String toString() {
			StringBuilder s = new StringBuilder();
			String N = System.getProperty("line.separator");

			if (objective != null) {
				s.append("min:");
				s.append(objective);
				s.append(';');
				s.append(N);
			}
			for (Constraint constraint : getConstraints()) {
				s.append(constraint);
				s.append(';');
				s.append(N);
			}
			for (Object variable : getVariables()) {
				Number lb = getVarLowerBound(variable);
				Number ub = getVarUpperBound(variable);

				if (lb != null || ub != null) {
					if (lb != null) {
						s.append(lb + " <= ");
					}
					s.append(variable);
					if (ub != null) {
						s.append(" <= " + ub);
					}
					s.append(';');
					s.append(N);
				}
			}

			List<Object> booleanVars = new ArrayList<>();
			List<Object> intVars = new ArrayList<>();
			for (Object variable : getVariables()) {
				if (getVarType(variable) == VarType.BOOL) {
					booleanVars.add(variable);
				} else if (getVarType(variable) == VarType.INT) {
					intVars.add(variable);
				}
			}

			// integer variables
			s.append("int ");
			for (Iterator<Object> iterator = intVars.iterator(); iterator.hasNext();) {
				Object var = iterator.next();
				s.append(var);
				if (iterator.hasNext()) {
					s.append(',');
				}
			}
			s.append(';');
			s.append(N);

			// boolean variables
			s.append("binary ");
			for (Iterator<Object> iterator = booleanVars.iterator(); iterator.hasNext();) {
				Object var = iterator.next();
				s.append(var);
				if (iterator.hasNext()) {
					s.append(',');
				}
			}
			s.append(';');
			s.append(N);

			return s.toString();
		}

	}

	private final Problem problem;
	private final Linear objective;
	private Multimap<String, Constraint> nullVariables;

	public ILPProblemImpl() {
		super();
		problem = new DebugableProblem();
		objective = new Linear();
	}

	public Result solve(Solver solver) {

		problem.setObjective(objective, OptType.MIN);

		if (problem.getConstraintsCount() == 0 && problem.getObjective().size() == 0) {
			return null;
		}
		Stopwatch watch = Stopwatch.createStarted();
		Result result = solver.solve(problem);
		if (watch.elapsed(TimeUnit.SECONDS) > 10) {
			System.out.println("Slow MILP solving! Took "+watch.elapsed(TimeUnit.SECONDS));			
		}
		return result;
	}

	public void addObjective(double coefficient, String varName) {
		objective.add(coefficient, varName);
	}

	public int getObjectiveSize() {
		return objective.size();
	}

	public void addORConstraints(Linear lhs, Operator operator, Number rhs) {
		problem.add(new Constraint(CONSTRAINT_NAME, lhs, operator, rhs));
	}

	public void addGuardConstraints(Linear lhs, Operator operator, Number rhs) {
		problem.add(new Constraint(CONSTRAINT_NAME, lhs, operator, rhs));
	}

	public void addAssignmentConstraint(Linear lhs, Operator operator, Number rhs) {
		problem.add(new Constraint(CONSTRAINT_NAME, lhs, operator, rhs));
	}

	public int getConstraintsCount() {
		return problem.getConstraintsCount();
	}

	public void setVarLowerBound(String variable, double value) {
		problem.setVarLowerBound(variable, value);
	}

	public Number getVarLowerBound(String varName) {
		return problem.getVarLowerBound(varName);
	}

	public void setVarUpperBound(String variable, double value) {
		problem.setVarUpperBound(variable, value);
	}

	public Number getVarUpperBound(String varName) {
		return problem.getVarUpperBound(varName);
	}

	public void setVarType(String varName, Class<?> type) {
		problem.setVarType(varName, type);
	}

	public Class<?> getVarType(String varName) {
		VarType varType = problem.getVarType(varName);
		switch (varType) {
			case BOOL :
				return Boolean.class;
			case INT :
				return Integer.class;
			case REAL :
				return Double.class;
			default :
				return Double.class;
		}
	}

	public String toString() {
		return problem.toString();
	}

	public void addNullConstraint(String ilpVarName, Linear lhs, Operator operator, Number rhs) {
		if (nullVariables == null) {
			nullVariables = HashMultimap.create();
		}
		Constraint constraint = new Constraint(CONSTRAINT_NAME, lhs, operator, rhs);
		nullVariables.put(ilpVarName, constraint);
	}

	public double writeNullConstraint(String ilpVarName, boolean wasNull) {

		for (Constraint constraintTemplate : nullVariables.get(ilpVarName)) {

			double rhs = (Double) constraintTemplate.getRhs();
			assert Double.isInfinite(rhs);
			// NEGATIVE_INFINITY is x==NULL 
			// POSITIVE_INFINITY is x!=NULL
			if (wasNull) {

				// x was not written 

				if (rhs == Double.NEGATIVE_INFINITY) {
					// NULL == NULL
					if (constraintTemplate.getLhs().size() > 0) {
						// is connected to OR, thus, the OR variable may safely be 0 
						addGuardConstraints(constraintTemplate.getLhs(), Operator.LE, 0.0d);
					}
					// else ignore it: X was NULL and not connected to other constraints -> always TRUE
				} else {
					// NULL != NULL
					if (constraintTemplate.getLhs().size() > 0) {
						// is connected to OR, thus, the OR variable needs to be 1 to fulfill guard
						addGuardConstraints(constraintTemplate.getLhs(), Operator.GE, 1.0d);
					} else {
						// x was NULL, thus, always FALSE
						return Double.POSITIVE_INFINITY;
					}
				}

			} else {

				// x was written 

				if (rhs == Double.NEGATIVE_INFINITY) {
					// !NULL == NULL
					if (constraintTemplate.getLhs().size() > 0) {
						// is connected to OR, thus, the OR variable needs to be 1 to fulfill guard 
						addGuardConstraints(constraintTemplate.getLhs(), Operator.GE, 1.0d);
					} else {
						// x was not NULL, thus, always FALSE
						return Double.POSITIVE_INFINITY;
					}
				} else {
					// !NULL != NULL
					if (constraintTemplate.getLhs().size() > 0) {
						// is connected to OR, thus, the OR variable may safely be 0 
						addGuardConstraints(constraintTemplate.getLhs(), Operator.LE, 0.0d);
					}
					// else ignore it: X was not NULL and not connected to other constraints -> always TRUE
				}
			}
		}
		return 0.0d;
	}

	public boolean hasNullConstraint(String ilpVariable) {
		if (nullVariables == null) {
			return false;
		}
		return nullVariables.containsKey(ilpVariable);
	}

}