/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.export;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.NumberFormat;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UIExportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.Alignment.AlignmentStep;
import org.processmining.plugins.DataConformance.ResultReplay;
import org.processmining.plugins.DataConformance.framework.ExecutionStep;
import org.processmining.plugins.balancedconformance.result.AlignmentCollection;

import au.com.bytecode.opencsv.CSVWriter;

import com.google.common.collect.Lists;

/**
 * Exports a {@link AlignmentCollection} as CSV, including meta-data such as
 * fitness, and, possibly, the attributes of the event log.
 * 
 * @author F. Mannhardt
 */
@Plugin(name = "Export Alignments as CSV (.csv)", level = PluginLevel.Regular, returnLabels = {}, returnTypes = {}, parameterLabels = {
		"ResultReplayPetriNetWithData", "File" }, userAccessible = true)
@UIExportPlugin(description = "Export Alignments as CSV (.csv)", extension = "csv")
public class ExportAsCSVBalancedReplayResult {

	private final NumberFormat nf = NumberFormat.getInstance();
	private final NumberFormat nfInt = NumberFormat.getInstance();

	@PluginVariant(variantLabel = "Export Alignments as CSV (.csv)", requiredParameterLabels = { 0, 1 })
	public void exportAsCSV(final PluginContext context, AlignmentCollection res, File file) throws FileNotFoundException, IOException {
		nf.setMinimumFractionDigits(8);
		nf.setMaximumFractionDigits(8);

		nfInt.setMinimumFractionDigits(0);
		nfInt.setMaximumFractionDigits(0);

		exportToFile(context, res, file);
	}

	private void exportToFile(PluginContext context, AlignmentCollection alignments, File file)
			throws FileNotFoundException, IOException {

		try (FileOutputStream out = new FileOutputStream(file)) {
			try (CSVWriter writer = new CSVWriter(new OutputStreamWriter(out, "UTF-8"), ',', '"', '\\')) {

				// Compile header

				Set<String> variablesSet = collectVariables(alignments);
				Set<String> attributesSet;
				if (alignments instanceof ResultReplay) {
					attributesSet = collectLogAttributes(variablesSet, (ResultReplay) alignments);
				} else {
					attributesSet = Collections.<String>emptySet();
				}

				List<String> headerList = Lists.newArrayList("traceName", "alignmentFitness", "alignmentCost", "moveNr", "moveType", "processActivity", "logActivity");

				Iterator<String> varIter = variablesSet.iterator();
				while (varIter.hasNext()) {
					String varName = varIter.next();
					headerList.add("process:var:" + varName);
					headerList.add("log:var:" + varName);
				}			

				// Add additional attributes 

				Iterator<String> attributesIter = attributesSet.iterator();
				while (attributesIter.hasNext()) {
					String attributeName = attributesIter.next();
					headerList.add("log:attribute:" + attributeName);
				}

				writer.writeNext(headerList.toArray(new String[headerList.size()]), false);

				// Write alignments

				String[] nextLine = new String[headerList.size()];

				for (Alignment alignment : alignments) {
					
					int stepNr = -1;

					for (AlignmentStep step : alignment) {

						int i = -1;

						nextLine[++i] = alignment.getTraceName() != null ? alignment.getTraceName() : "";						
						nextLine[++i] = nf.format(alignment.getFitness());
						nextLine[++i] = nf.format(alignment.getCost());						
						nextLine[++i] = nfInt.format(++stepNr);
						nextLine[++i] = step.getType().name();
						nextLine[++i] = step.getProcessView().getActivity();
						nextLine[++i] = step.getLogView().getActivity();

						varIter = variablesSet.iterator();

						while (varIter.hasNext()) {
							String var = varIter.next();
							Object value = step.getProcessView().get(var);
							if (value != null) {
								nextLine[++i] = value.toString();
							} else {
								nextLine[++i] = "";
							}
							value = step.getLogView().get(var);
							if (value != null) {
								nextLine[++i] = value.toString();
							} else {
								nextLine[++i] = "";
							}
						}

						// Add additional attributes, if XEvent is found on LogView

						if (step.getLogView() != null && step.getLogView().getActivityObject() instanceof XEvent) {
							attributesIter = attributesSet.iterator();
							while (attributesIter.hasNext()) {
								String attributeKey = attributesIter.next();
								XAttribute attribute = ((XEvent) step.getLogView().getActivityObject()).getAttributes()
										.get(attributeKey);
								if (attribute != null) {
									nextLine[++i] = attribute.toString();
								} else {
									nextLine[++i] = "";
								}
							}
						}

						// Remove old values, if not yet overwritten
						if (i + 1 < nextLine.length) {
							for (int j = i + 1; j < nextLine.length; j++) {
								nextLine[j] = "";
							}
						}

						writer.writeNext(nextLine);

					}
				}
				writer.flush();
			}
		}

	}

	private SortedSet<String> collectLogAttributes(Set<String> variablesSet, ResultReplay alignments) {

		SortedSet<String> attributesSet = new TreeSet<>();

		for (XTrace t : alignments.getAlignedLog()) {
			for (XEvent e : t) {
				attributesSet.addAll(e.getAttributes().keySet());
			}
		}

		// Remove those that are already covered in the variables

		for (String variable : variablesSet) {
			String attributeAlreadyCovered = alignments.getVariableMapping().get(variable);
			attributesSet.remove(attributeAlreadyCovered);
		}

		return attributesSet;

	}

	private SortedSet<String> collectVariables(AlignmentCollection alignments) {

		SortedSet<String> variablesSet = new TreeSet<>();

		for (Alignment a : alignments) {
			for (ExecutionStep e : a.getProcessTrace()) {
				variablesSet.addAll(e.keySet());
			}
		}

		return variablesSet;
	}

}
