/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.observer;

import java.util.Collection;
import java.util.logging.Level;

import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.balancedconformance.dataflow.exception.UnreliableDataAlignmentException;
import org.processmining.plugins.balancedconformance.result.BalancedDataAlignmentState;
import org.processmining.plugins.balancedconformance.result.DataAlignedTrace;

/**
 * Observer class that provides information during the compution of alignments
 * 
 * @author F. Mannhardt
 *
 */
public interface DataConformancePlusObserver {

	public interface ImpossibleTrace {

		/**
		 * @return the original trace
		 */
		XTrace getTrace();

		/**
		 * @return an error message
		 */
		String getMessage();

		UnreliableDataAlignmentException.UnreliableReason getReason();

	}

	/**
	 * Called at the start of the alignment
	 * 
	 * @param numExpectedResults
	 */
	void startAlignment(int numExpectedResults);

	/**
	 * Called for each new trace
	 * 
	 * @param traceIndex
	 * @param trace
	 */
	void processingNewTrace(int traceIndex, XTrace trace);

	/**
	 * An optimal alignment was found
	 * 
	 * @param traceIndex
	 * @param trace
	 * @param dataAlignment
	 * @param partialDataAlignmentsNeeded
	 * @param cacheHit
	 * @param cacheSize
	 * @param queuedStates
	 * @param dataStateCount
	 * @param usedTime
	 */
	void foundOptimalAlignment(int traceIndex, XTrace trace, DataAlignedTrace dataAlignment,
			int partialDataAlignmentsNeeded, int cacheHit, int cacheSize, long queuedStates, long dataStateCount,
			long usedTime);

	/**
	 * The fitness was calculated based on the obtained optimal alignment
	 * 
	 * @param traceIndex
	 * @param trace
	 * @param resultState
	 */
	void calculatedFitness(int traceIndex, XTrace trace, BalancedDataAlignmentState resultState);

	/**
	 * The alignment computation was slow for this trace
	 * 
	 * @param traceIndex
	 * @param trace
	 * @param dataAlignment
	 * @param usedTime
	 */
	void slowDataAlignmentDetected(int traceIndex, XTrace trace, DataAlignedTrace dataAlignment, long usedTime);

	/**
	 * No alignment could be computed for this trace. Possibly because there the
	 * cost exceeded the configured threshold.
	 * 
	 * @param traceIndex
	 * @param currentTrace
	 */
	void unreliableAlignmentDetected(int traceIndex, XTrace currentTrace);

	/**
	 * At the end of the alignment computation this method is called with the
	 * (possibly empty) list of traces for which no alignment could be obtained.
	 * <p>
	 * This is a convenience method to be used instead of
	 * {@link #unreliableAlignmentDetected(int, XTrace)}.
	 * 
	 * @param impossibleTraces
	 */
	void foundImpossibleAlignments(Collection<ImpossibleTrace> impossibleTraces);

	/**
	 * Use {@link #log(Level, String)}
	 */
	@Deprecated
	void log(String message);

	/**
	 * Use {@link #log(Level, String)}
	 */
	@Deprecated
	void log(String message, Throwable e);

	/**
	 * 
	 * 
	 * @param level
	 * @param message
	 */
	void log(Level level, String message);

	/**
	 * The alignment computation is finished
	 */
	void finishedAlignment();

}
