/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 */
package org.processmining.plugins.balancedconformance.ui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import nl.tue.astar.AStarThread.QueueingModel;
import nl.tue.astar.Tail;
import nl.tue.astar.util.FastLookupPriorityQueue;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.ui.ConfigurationUIHelper;
import org.processmining.framework.connections.Connection;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionID;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.models.connections.petrinets.EvClassLogPetrinetConnection;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.plugins.DataConformance.DataAlignment.PetriNet.ReplayableTransition;
import org.processmining.plugins.DataConformance.GUI.MappingPanel;
import org.processmining.plugins.DataConformance.GUI.VariableMatchCostPanel;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.astar.petrinet.impl.AbstractPDelegate;
import org.processmining.plugins.balancedconformance.cache.DataViewBuilder;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.config.DataConformancePlusConfiguration;
import org.processmining.plugins.balancedconformance.controlflow.adapter.AlignmentAdapter.PriorityQueueFactory;
import org.processmining.plugins.balancedconformance.controlflow.adapter.SearchMethod;
import org.processmining.plugins.balancedconformance.controlflow.override.queues.duplicates.BFDuplicatePriorityQueue;
import org.processmining.plugins.balancedconformance.ui.override.StrippedDownPNReplayerUI;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.petrinet.replayer.algorithms.costbasedcomplete.CostBasedCompleteParam;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;

/**
 * Utilities for getting the configuration from the user via the
 * {@link UIPluginContext}.
 * 
 * @author F. Mannhardt
 * 
 */
public class DataConformanceConfigUIUtils {

	public static class VariableBounds {

		public Map<String, Object> upperBounds;
		public Map<String, Object> lowerBounds;

	}

	public static TransEvClassMapping queryActivityMapping(UIPluginContext context, DataPetriNet net, XLog log)
			throws UserCancelledException {
		try {
			EvClassLogPetrinetConnection conn = getFirstConnection(context.getConnectionManager(),
					EvClassLogPetrinetConnection.class, net, log);
			return (TransEvClassMapping) conn.getObjectWithRole(EvClassLogPetrinetConnection.TRANS2EVCLASSMAPPING);
		} catch (ConnectionCannotBeObtained e) {
			return ConfigurationUIHelper.queryActivityEventClassMapping(context, net, log);
		}
	}

	public static BalancedProcessorConfiguration queryBalancedAlignmentConfig(UIPluginContext context)
			throws UserCancelledException {
		BalancedProcessorConfiguration config = new BalancedProcessorConfiguration();
		BalancedAlignmentConfigPanel balancedAlignmentConfigPanel = new BalancedAlignmentConfigPanel();
		InteractionResult balancedAlignmentResult = context
				.showConfiguration("Performance & Alignment-related Parameters", balancedAlignmentConfigPanel);
		if (balancedAlignmentResult == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			config.setActivateDataViewCache(balancedAlignmentConfigPanel.isDataViewCacheActivated());
			config.setUseOptimizations(balancedAlignmentConfigPanel.getIsUseOptimizations());
			config.setConcurrentThreads(balancedAlignmentConfigPanel.getConcurrentThreads());
			config.setSorting(balancedAlignmentConfigPanel.getSorting());
			config.setIlpSolver(balancedAlignmentConfigPanel.getILPSolver());
			config.setUsePartialDataAlignments(balancedAlignmentConfigPanel.getIsUsePartialDataAlignments());
			config.setVariablesUnassignedMode(balancedAlignmentConfigPanel.getUnassignedMode());
			config.setQueueingModel(balancedAlignmentConfigPanel.getQueueingModel());
			config.setMaxCostFactor(balancedAlignmentConfigPanel.getMaxCost());
			config.setUsePartialOrders(balancedAlignmentConfigPanel.getIsUsePartialOrders());

			config.setKeepControlFlowSearchSpace(balancedAlignmentConfigPanel.getIsKeepControlFlowSearchSpace());
			config.setKeepDataFlowSearchSpace(balancedAlignmentConfigPanel.getIsKeepDataFlowSearchSpace());
			config.setSearchMethod(balancedAlignmentConfigPanel.getSearchMethod());
			config.setMaxQueuedStates(balancedAlignmentConfigPanel.getMaxQueuedStates());
			return config;
		}
	}

	@Deprecated
	public static PriorityQueueFactory createQueueFactory(QueueingModel queueingModel, SearchMethod method) {
		return new PriorityQueueFactory() {
			
			public FastLookupPriorityQueue newInstance(DataViewBuilder dataViewBuilder,
					AbstractPDelegate<? extends Tail> delegate) {
				return new BFDuplicatePriorityQueue(1024);
			}
		}; 
	}

	@SuppressWarnings("unchecked")
	private static <T extends Connection> T getFirstConnection(ConnectionManager connectionManager,
			Class<T> connectionType, Object... objects) throws ConnectionCannotBeObtained {
		for (ConnectionID connID : connectionManager.getConnectionIDs()) {
			Connection c = connectionManager.getConnection(connID);
			if (((connectionType == null) || connectionType.isAssignableFrom(c.getClass()))
					&& c.containsObjects(objects)) {
				return (T) c;
			}
		}
		throw new ConnectionCannotBeObtained("Connection not found", connectionType, objects);
	}

	public static void queryControlFlowAlignmentConfig(UIPluginContext context, PetrinetGraph net, XLog log,
			DataConformancePlusConfiguration config, XEventClasses eventClasses) throws UserCancelledException {
		CostBasedCompleteParam userConfig = new StrippedDownPNReplayerUI().getConfiguration(context, net, log, eventClasses);
		if (userConfig != null) {
			config.setInitialMarking(userConfig.getInitialMarking());
			config.setFinalMarkings(userConfig.getFinalMarkings());
			config.setMapEvClass2Cost(userConfig.getMapEvClass2Cost());
			config.setMapTrans2Cost(userConfig.getMapTrans2Cost());
		} else {
			throw new UserCancelledException();
		}
	}

	public static void queryControlFlowAlignmentConfig(final UIPluginContext context, PetrinetGraph net, XLog log,
			DataConformancePlusConfiguration config) throws UserCancelledException {
		queryControlFlowAlignmentConfig(context, net, log, config,
				XUtils.createEventClasses(config.getActivityMapping().getEventClassifier(), log));
	}

	/**
	 * @param context
	 * @param net
	 * @param summary
	 * @return mapping from variable name in DPN-net to variable name in event
	 *         log
	 * 
	 * @throws UserCancelledException
	 */
	public static MappingPanel<String, String> queryVariableMapping(final UIPluginContext context,
			final DataPetriNet net, final Set<String> eventEttributeKeys) throws UserCancelledException {

		TreeSet<String> processAttributes = new TreeSet<>();
		for (DataElement variable : net.getVariables()) {
			processAttributes.add(variable.getVarName());
		}

		MappingPanel<String, String> mapVariablePanel = new MappingPanel<>(processAttributes, eventEttributeKeys);
		InteractionResult result = context.showConfiguration("Setup Variable Mapping", mapVariablePanel);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		}
		return mapVariablePanel;
	}

	public static VariableMatchCosts queryVariableMatchCosts(final UIPluginContext context,
			final TransEvClassMapping activityMapping, final Map<String, String> variableMappingWithUnmapped)
			throws UserCancelledException {

		HashMap<ReplayableTransition, XEventClass> activityMapping2 = new HashMap<>();
		for (Entry<Transition, XEventClass> entry : activityMapping.entrySet()) {
			activityMapping2.put(new ReplayableTransition(entry.getKey()), entry.getValue());
		}

		VariableMatchCostPanel<XEventClass> variablePanel = new VariableMatchCostPanel<>(activityMapping2,
				variableMappingWithUnmapped);

		InteractionResult result = context.showConfiguration("Cost of Deviations in Data Perspective", variablePanel);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		}

		return variablePanel.getCosts();
	}

	public static VariableBounds queryVariableBounds(final UIPluginContext context, final DataPetriNet net,
			final XLog log, final Map<String, String> variableMapping) throws UserCancelledException {
		context.log(
				"Derive appropriate bounds for the ILP problems based on the values seen in the Event Log. This iterates through the whole Event log and may take a while ...");
		CheckVariableBoundsPanel variableBoundsPanel = new CheckVariableBoundsPanel(net, log, variableMapping);
		context.log("Finished deriving variable bounds.");

		InteractionResult result = context.showConfiguration("Configure Variable Bounds", variableBoundsPanel);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		}

		Map<String, Object> upperBounds = new HashMap<>();
		Map<String, Object> lowerBounds = new HashMap<>();

		for (DataElement elem : variableBoundsPanel.getDataElements()) {
			Object value = elem.getMinValue();
			if (value != null)
				lowerBounds.put(elem.getVarName(), value);
			value = elem.getMaxValue();
			if (value != null)
				upperBounds.put(elem.getVarName(), value);
		}

		VariableBounds variableBounds = new VariableBounds();
		variableBounds.upperBounds = upperBounds;
		variableBounds.lowerBounds = lowerBounds;
		return variableBounds;
	}

	/**
	 * Use
	 * {@link #queryDataAlignmentConfig(UIPluginContext, DataPetriNet, XLog, TransEvClassMapping, DataConformancePlusConfiguration)}
	 * 
	 * @throws UserCancelledException
	 */
	@Deprecated
	public static void queryDataAlignmentConfig(final UIPluginContext context, DataPetriNet net, XLog log,
			XLogInfo logInfo, TransEvClassMapping activityMapping, DataConformancePlusConfiguration config)
			throws UserCancelledException {
		queryDataAlignmentConfig(context, net, log, activityMapping, config);
	}

	public static void queryDataAlignmentConfig(final UIPluginContext context, DataPetriNet net, XLog log,
			TransEvClassMapping activityMapping, DataConformancePlusConfiguration config)
			throws UserCancelledException {
		Set<String> attributeNames;
		if (log.getInfo(activityMapping.getEventClassifier()) != null) {
			XLogInfo info = log.getInfo(activityMapping.getEventClassifier());
			attributeNames = ImmutableSet.copyOf(info.getEventAttributeInfo().getAttributeKeys());
		} else {
			attributeNames = XUtils.getEventAttributeKeys(log);
		}

		MappingPanel<String, String> variableMapping = queryVariableMapping(context, net, attributeNames);
		config.setVariableMapping(variableMapping.getMapping(false));

		VariableMatchCosts variableCosts = queryVariableMatchCosts(context, activityMapping,
				variableMapping.getMapping(true));
		config.setVariableCost(variableCosts);

		VariableBounds variableBounds = queryVariableBounds(context, net, log, config.getVariableMapping());

		config.setUpperBounds(variableBounds.upperBounds);
		config.setLowerBounds(variableBounds.lowerBounds);
	}


	public static void queryControlFlowCost(final UIPluginContext context, DataPetriNet net, XLog log,
			DataConformancePlusConfiguration config) throws UserCancelledException {
		queryControlFlowCost(context, net, log, config, 1, 1);
	}

	public static void queryControlFlowCost(final UIPluginContext context, DataPetriNet net, XLog log,
			DataConformancePlusConfiguration config, int defaultCostMoveOnLog, int defaultCostMoveOnModel)
			throws UserCancelledException {

		// populate event classes
		XEventClassifier classifier = config.getActivityMapping().getEventClassifier();
		XLogInfo info = log.getInfo(classifier);
		if (info == null) {
			info = XLogInfoFactory.createLogInfo(log, classifier);
		}
		XEventClasses eventClassesName = info.getEventClasses();
		List<XEventClass> evClasses = new ArrayList<>(eventClassesName.getClasses());
		evClasses.add(config.getActivityMapping().getDummyEventClass());
		ArrayList<Transition> transitionList = Lists.newArrayList(net.getTransitions());
		Collections.sort(evClasses, new Comparator<XEventClass>() {

			public int compare(XEventClass o1, XEventClass o2) {
				return o1.getId().compareTo(o2.getId());
			}
		});
		Collections.sort(transitionList, new Comparator<Transition>() {

			public int compare(Transition o1, Transition o2) {
				return o1.getLabel().compareTo(o2.getLabel());
			}
		});

		ControlFlowCostUI controlFlowCostUI = new ControlFlowCostUI(transitionList, evClasses,
				config.getActivityMapping(), defaultCostMoveOnLog, defaultCostMoveOnModel);

		InteractionResult result = context.showConfiguration("Cost of Deviations in Control-Flow Perspective",
				controlFlowCostUI);
		if (result == InteractionResult.CANCEL) {
			throw new UserCancelledException();
		} else {
			config.setMapEvClass2Cost(controlFlowCostUI.getMapEvClassToCost());
			config.setMapTrans2Cost(controlFlowCostUI.getTransitionWeight());
		}
	}

}
